import numpy as np
import pickle
import multiprocessing as mp
import matplotlib.pyplot as plt
from qutip import * 

import scipy.sparse as spr
import block_diagonalize
import time

def run_fun(Delta):
    Grange = np.geomspace(0.05, 50, 51)*Delta
    c_ops = [np.sqrt(eta)*a**2]#, np.sqrt(beta)*a.dag()*a]

    num_even     = np.zeros(Grange.size)
    num_odd      = np.zeros(Grange.size)
    gap_even     = np.zeros(Grange.size, dtype='complex')
    gap_odd      = np.zeros(Grange.size, dtype='complex')
    gap_even_odd = np.zeros(Grange.size, dtype='complex')
    gap_odd_even = np.zeros(Grange.size, dtype='complex')

    for g in range(Grange.size):
        H = -Delta*a.dag()*a + Grange[g]/2 *(a.dag()**2 +a**2) +U/2 *a.dag()**2*a**2
        LL = liouvillian(H, c_ops)

        P, block_bfs, block_sizes = block_diagonalize.PermMat(LL)
        bd_L = np.dot(P, np.dot(LL.data, np.transpose(P)))
        num_blocks, blocks_list, bl_indices = block_diagonalize.get_blocks(bd_L)   

        done=False
        for i in range(int(len(blocks_list))):
            block = blocks_list[i]
            evals, evecs = Qobj(block).eigenstates()    

            ss_block_form=evecs[-1]
            evec2 = spr.dok_matrix((LL.shape[0], 1), dtype=np.complex)
            evec2[bl_indices[i]:bl_indices[i]+blocks_list[i].shape[0]] = ss_block_form.data
            evec2 = evec2.tocsr()
            evec2=np.dot(np.transpose(P),evec2)
            ss= Qobj(evec2, dims=[LL.dims[0], [1]])
            ss=vector_to_operator(ss)    
            if ss.tr()>0.05:
                ss=ss+ss.dag()
                ss/=(ss.tr())       
                if expect(parity, ss)>0.5:
                    num_even[g] = expect(ss, a.dag()*a)
                    gap_even[g] = evals[-2]
                else:
                    num_odd[g] = expect(ss, a.dag()*a)                
                    gap_odd[g] = evals[-2]
            elif done==False:
                gap_even_odd[g] = np.real(evals[-1])+1.j*np.abs(np.imag(evals[-1]))
                gap_odd_even[g] = np.real(evals[-1])-1.j*np.abs(np.imag(evals[-1]))       
                done=True
                
    return [Grange, num_even, num_odd, gap_even, gap_odd, gap_even_odd, gap_odd_even]


if __name__=="__main__":
    #Setup
    a=destroy(21)

    parity=1.j*np.pi*a.dag()*a
    parity=parity.expm()

    U   = 1
    eta = 1
    
    NCPU = mp.cpu_count()
    Deltarange = np.geomspace(0.01, 100, NCPU)
    
    #perform parallel evaluation
    t1 = time.time()
    pool = mp.Pool(NCPU)
    results = pool.map(run_fun,Deltarange)
    print(time.time()-t1)

    with open("GoverDelta_scan.pckl", "wb") as fp:
        pickle.dump([Deltarange, results], fp)
        
        
        
        
        